#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>

#include "util.h"
#include "doc.h"
#include "slda.h"

strct_slda* alloc_slda(int K, int V) {

	strct_slda *slda;

	printf("sizeof(strct_slda) = %d\n", (int)(sizeof(strct_slda)));

	assert(slda = (struct strct_slda*)malloc(sizeof(struct strct_slda)));
	slda->K = K;
	slda->V = V;
	slda->alpha = gsl_vector_alloc(K);
	assert(slda->mu = (gsl_vector**)malloc(K * sizeof(gsl_vector*)));
	for(int k = 0; k < K; k++) {
		slda->mu[k] = gsl_vector_alloc(V);
	}
	slda->eta = gsl_vector_alloc(K);	

	return slda;

}

void free_slda(strct_slda* slda) {

	int K = slda->K;
	int V = slda->V;

	gsl_vector_free(slda->alpha);
	for(int k = 0; k < K; k++)
		gsl_vector_free(slda->mu[k]);
	gsl_vector_free(slda->eta);

}

void init_slda(strct_slda* slda, gsl_rng *rng) {

	int K = slda->K;
	int V = slda->V;

	gsl_vector_set_all(slda->alpha, DEFAULT_ALPHA);

	for(int k = 0; k < K; k++) {
		double sum = 0;
		for(int v = 0; v < V; v++) {
			slda->mu[k]->data[v] = gsl_ran_flat(rng, 0, 1);
			sum += slda->mu[k]->data[v];
		}
		for(int v = 0; v < V; v++)
			slda->mu[k]->data[v] = (slda->mu[k]->data[v] + 1e-3/V) / (sum + 1e-3);
	}

	for(int k = 0; k < K; k++) {
		slda->eta->data[k] = gsl_ran_gaussian(rng, 1.0);
	}

}

void write_slda(strct_slda* slda, FILE *f) {

	fprintf(f, "%d %d\n", slda->K, slda->V);
	for(int k = 0; k < slda->K; k++)
		fprintf(f, "%lf ", slda->alpha->data[k]);
	fprintf(f, "\n");
	for(int k = 0; k < slda->K; k++) {
		for(int i = 0; i < slda->V; i++)
			fprintf(f, "%lf ", slda->mu[k]->data[i]);
		fprintf(f, "\n");
	}
	for(int k = 0; k < slda->K; k++)
		fprintf(f, "%lf ", slda->eta->data[k]);
	fprintf(f, "\n");

}

strct_slda* read_slda(FILE *f) {

	int K, V;

	fscanf(f, "%d%d", &K, &V);
	strct_slda* slda = alloc_slda(K, V);
	for(int k = 0; k < K; k++)
		fscanf(f, "%lf", &(slda->alpha->data[k]));
	for(int k = 0; k < K; k++)
		for(int i = 0; i < V; i++)
			fscanf(f, "%lf", &(slda->mu[k]->data[i]));
	for(int k = 0; k < K; k++)
		fscanf(f, "%lf", &(slda->eta->data[k]));

	// smoothing
	for(int k = 0; k < K; k++)
		for(int v = 0; v < V; v++)
			slda->mu[k]->data[v] = (slda->mu[k]->data[v] + 1e-3/V) / (1.0 + 1e-3);

	return slda;

}

strct_doc* gen_docs_from_slda(strct_slda* slda, int ndocs, int M, double sigma, gsl_rng* rng) {

	int K = slda->K;
	int V = slda->V;

	strct_doc* docs;
	assert(docs = (strct_doc*)malloc(ndocs * sizeof(strct_doc)));

	unsigned int* count;
	assert(count = (unsigned int*)malloc(V * sizeof(int)));
	double* distrb;
	assert(distrb = (double*)malloc(V * sizeof(double)));

	for(int i = 0; i < ndocs; i++) {

		docs[i].word_count = M;
		init_hz(&(docs[i]), K, rng);

		// Step 1: generate the topic mixing vector
		gsl_ran_dirichlet(rng, K, slda->alpha->data, docs[i].h->data);

		// Step 2: generate the words
		memset(distrb, 0, V * sizeof(double));
		for(int j = 0; j < V; j++)
			for(int k = 0; k < K; k++)
				distrb[j] += docs[i].h->data[k] * slda->mu[k]->data[j];
		// check
		double sum = 0;
		for(int j = 0; j < V; j++) sum += distrb[j];
		assert(fabs(sum - 1) < MINEPS);

		memset(count, 0, V * sizeof(int));
		gsl_ran_multinomial(rng, V, M, distrb, count);

		docs[i].item_count = 0;
		docs[i].word_count = 0;
		for(int j = 0; j < V; j++)
			if (count[j] > 0) {
				docs[i].item_count ++;
				docs[i].word_count += count[j];
			}
		assert(docs[i].word_count == M);
		assert(docs[i].index = (int*)malloc(docs[i].item_count * sizeof(int)));
		assert(docs[i].count = (int*)malloc(docs[i].item_count * sizeof(int)));
		int p = 0;
		for(int j = 0; j < V; j++)
			if (count[j] > 0) {
				docs[i].index[p] = j;
				docs[i].count[p] = count[j];
				p ++;
			}
		assert(p == docs[i].item_count);

		// Step 3: generate the response variable y
		docs[i].y = 0;
		gsl_blas_ddot(slda->eta, docs[i].h, &(docs[i].y));
		docs[i].y += gsl_ran_gaussian(rng, sigma);

	}

	free(distrb);
	free(count);
	
	return docs;

}

void regularize_eta(gsl_vector* eta, int K, strct_doc* docs, int ndocs, double ref_mean, double ref_variance) {

	double mean = 0;
	double var = 0;
	gsl_vector* barz = gsl_vector_alloc(K);
	for(int i = 0; i < ndocs; i++) {
		gsl_vector_set_all(barz, 0);
		for(int j = 0; j < docs[i].word_count; j++)
			barz->data[docs[i].z[j]] += 1.0 / docs[i].word_count;
		double yhat = 0;
		gsl_blas_ddot(barz, eta, &yhat);
		mean += yhat;
		var += SQR(yhat);
	}
	gsl_vector_free(barz);

	mean /= ndocs;
	var = var / ndocs - SQR(mean);

	printf("mean = %lf, var = %lf, ref_mean = %lf, ref_var = %lf\n", mean, var, ref_mean, ref_variance);

	// adjust the mean
	for(int k = 0; k < K; k++)
		eta->data[k] -= (mean - ref_mean) / K;

	// adjust the variance
	for(int k = 0; k < K; k++)
		eta->data[k] /= sqrt(var / ref_variance);

}
